/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout.auto;

import static de.cau.cs.kieler.kiml.options.Direction.RIGHT;
import static de.cau.cs.kieler.kiml.options.LayoutOptions.DIRECTION;
import static de.cau.cs.kieler.kiml.options.LayoutOptions.INTERACTIVE;
import static de.cau.cs.kieler.kiml.options.LayoutOptions.PORT_SIDE;
import static de.cau.cs.kieler.kiml.util.KimlUtil.createInitializedEdge;
import static de.cau.cs.kieler.kiml.util.KimlUtil.createInitializedLabel;
import static de.cau.cs.kieler.kiml.util.KimlUtil.createInitializedNode;
import static de.cau.cs.kieler.kiml.util.KimlUtil.createInitializedPort;
import static de.cau.cs.kieler.klay.layered.properties.InteractiveReferencePoint.TOP_LEFT;
import static de.cau.cs.kieler.klay.layered.properties.Properties.FEEDBACK_EDGES;
import static de.cau.cs.kieler.klay.layered.properties.Properties.INTERACTIVE_REFERENCE_POINT;
import static java.lang.Thread.sleep;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase.isUpdateEnabled;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase.setUpdateEnabled;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.addBendPointToConnection;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodeLayoutData;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setStickyConnectorLayoutData;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.Map.Entry;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.BasicEMap;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.EditPartViewer;
import org.eclipse.swt.graphics.GC;
import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.base.model.layout.impl.PointsImpl;
import org.fortiss.tooling.base.ui.ToolingBaseUIActivator;
import org.fortiss.tooling.base.ui.editor.DiagramEditorBase;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.extension.IModelEditor;
import org.fortiss.tooling.kernel.ui.service.IModelEditorBindingService;

import de.cau.cs.kieler.core.alg.BasicProgressMonitor;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KLabel;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.AbstractLayoutProvider;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KPoint;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.PortSide;
import de.cau.cs.kieler.klay.layered.LayeredLayoutProvider;

/**
 * <p>
 * {@link IAutoLayouter} based on <a
 * href="https://rtsys.informatik.uni-kiel.de/confluence/display/KIELER/Home">KIELER framework</a>.
 * </p>
 * <P>
 * Approach:
 * <ul>
 * <li>Generate a KIELER graph from the input model</li>
 * <li>Apply the KIELER <i>layered layout algorithm</i></li>
 * <li>Transfer generated layout to input model
 * </ul>
 * 
 * @author offtermatt, barner
 */
public class KielerAutoLayouter implements IAutoLayouter {

	/** Mapping from {@link IHierarchicElement}s to {@link KNode}s. */
	private EMap<IHierarchicElement, KNode> modelElementsToKNodes;

	/** Mapping from outbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> outboundConnectorsToKPorts;

	/** Mapping from inbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> inboundConnectorsToKPorts;

	/** Used to map undirected {@link IConnector}s to {@link KPort}s */
	private EMap<IConnector, KPort> undirectedConnectorsToKPorts;

	/** Mapping from {@link IConnection}s to {@link KEdge}s. */
	private EMap<IConnection, KEdge> connectionsToKEdges;

	/** Mapping from global inbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> globalInboundConnectorsToKPorts;

	/** Mapping from global outbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> globalOutboundConnectorsToKPorts;

	/** {@inheritDoc} */
	@Override
	public void performAutoLayout(IHierarchicElement element) {
		boolean avUpdateEnabled = isUpdateEnabled();
		setUpdateEnabled(false);
		KNode rootNode = createKIELERGraph(element);

		KShapeLayout rootNodeLayout = rootNode.getData(KShapeLayout.class);
		rootNodeLayout.setProperty(DIRECTION, RIGHT);
		rootNodeLayout.setProperty(INTERACTIVE, true);
		rootNodeLayout.setProperty(FEEDBACK_EDGES, true);
		rootNodeLayout.setProperty(INTERACTIVE_REFERENCE_POINT, TOP_LEFT);

		AbstractLayoutProvider layoutProvider = new LayeredLayoutProvider();
		IKielerProgressMonitor progressMonitor = new BasicProgressMonitor();
		layoutProvider.doLayout(rootNode, progressMonitor);

		applyLayout(rootNode, element);

		setUpdateEnabled(avUpdateEnabled);
	}

	/**
	 * Truncates a given {@code float} value and decreases the value until it is a multiple of the
	 * grid size.
	 * 
	 * @param f
	 *            {@code float} to process
	 * 
	 * @return Given float value, truncated and aligned to the underlying grid.
	 */
	private int truncateSnap2Grid(float f) {
		return truncateSnap2Grid(f, false);
	}

	/**
	 * Truncates a given {@code float} value and aligns it the grid size.
	 * 
	 * @param f
	 *            {@code float} to process
	 * 
	 * @param enlarge
	 *            Whether to enlarge the given value while aligning to the grid.
	 * 
	 * @return Given float value, truncated and aligned to the underlying grid.
	 */
	private int truncateSnap2Grid(float f, boolean enlarge) {
		int rval = (int)f;

		final int deviationFromGrid = rval % DEFAULT_GRID_SIZE;
		if(deviationFromGrid > 0) {
			if(enlarge) {
				rval += DEFAULT_GRID_SIZE;
			}
		}
		rval -= deviationFromGrid;

		return rval;
	}

	/**
	 * Applies the layout defined in the {@link KNode} follows to given {@link IHierarchicElement}.
	 * 
	 * @param template
	 *            The {@link KNode} that stores the layout.
	 * @param element
	 *            The {@link IHierarchicElement} to which the layout should be applied.
	 */
	private void applyLayout(KNode template, IHierarchicElement element) {
		ITopLevelElement topLevelElement =
				IPersistencyService.getInstance().getTopLevelElementFor(element);
		topLevelElement.runAsCommand(() -> {
			filterBendPoints(element);

			for(IHierarchicElement child : element.getContainedElements()) {
				KNode kNode = modelElementsToKNodes.get(child);
				KShapeLayout kNodeLayout = kNode.getData(KShapeLayout.class);
				setNodeLayoutData((ILayoutedModelElement)child,
						truncateSnap2Grid(kNodeLayout.getXpos()),
						truncateSnap2Grid(kNodeLayout.getYpos()),
						truncateSnap2Grid(kNodeLayout.getWidth(), true),
						truncateSnap2Grid(kNodeLayout.getHeight(), true));
			}

			for(Entry<IConnection, KEdge> entry : connectionsToKEdges) {
				KEdge kEdge = entry.getValue();
				IConnection connection = entry.getKey();
				KEdgeLayout kEdgeLayout = kEdge.getData(KEdgeLayout.class);
				for(KPoint kPoint : kEdgeLayout.getBendPoints()) {
					addBendPointToConnection((ILayoutedModelElement)connection,
							truncateSnap2Grid(kPoint.getX()), truncateSnap2Grid(kPoint.getY()));
				}
			}

			setPositionsForConnectors(outboundConnectorsToKPorts);
			setPositionsForConnectors(inboundConnectorsToKPorts);
			setPositionsForGlobalConnectors(globalInboundConnectorsToKPorts);
			setPositionsForGlobalConnectors(globalOutboundConnectorsToKPorts);
		});
	}

	/**
	 * Applies the layout of the specified {@link KPort}s to their matching (global)
	 * {@link IConnector}s. Global means in this case that the {@link IConnector}s do not have a
	 * corresponding node, i.e. because they contain connections to a higher-level model.
	 * 
	 * @param globalConnectors
	 *            A bijection from {@link IConnector}s to {@link KPort}s
	 */
	private void setPositionsForGlobalConnectors(EMap<IConnector, KPort> globalConnectors) {
		for(Entry<IConnector, KPort> entry : globalConnectors) {
			IConnector connector = entry.getKey();
			KPort kPort = entry.getValue();
			KNode tempNode = kPort.getNode();
			KShapeLayout kLayout = tempNode.getData(KShapeLayout.class);
			setNodePosition((ILayoutedModelElement)connector, truncateSnap2Grid(kLayout.getXpos()),
					truncateSnap2Grid(kLayout.getYpos()));
		}
	}

	/**
	 * Applies the layout of the specified {@link KPort}s to their matching {@link IConnector}s. The
	 * {@link IConnector}s must have corresponding nodes, i.e. they can not contain an
	 * {@link IConnection} to a higher-level model.
	 * 
	 * @param connectors
	 *            A bijection from {@link IConnector}s to {@link KPort}s.
	 */
	private void setPositionsForConnectors(EMap<IConnector, KPort> connectors) {
		for(Entry<IConnector, KPort> entry : connectors) {
			KPort kPort = entry.getValue();
			IConnector connector = entry.getKey();
			KShapeLayout kLayout = kPort.getData(KShapeLayout.class);
			PortSide side = kLayout.getProperty(PORT_SIDE);
			EOrientation orientation = null;
			switch(side) {
				case NORTH:
					orientation = EOrientation.NORTH;
					break;
				case WEST:
					orientation = EOrientation.WEST;
					break;
				case EAST:
					orientation = EOrientation.EAST;
					break;
				case SOUTH:
					orientation = EOrientation.SOUTH;
					break;
			}

			setStickyConnectorLayoutData((ILayoutedModelElement)connector, orientation,
					truncateSnap2Grid((kLayout.getYpos())));
		}
	}

	/** Removes all bend-points in connections contained in the given {@link IHierarchicElement}. */
	private void filterBendPoints(IHierarchicElement element) {
		for(IConnection connection : element.getConnections()) {
			for(ILayoutData layoutData : ((ILayoutedModelElement)connection).getLayoutData()) {
				if(layoutData instanceof Points || layoutData instanceof PointsImpl) {
					((Points)layoutData).getPoints().clear();
				}
			}
		}
	}

	/**
	 * Opens the given {@link IHierarchicElement} in an editor. Returns the editor in case it is a
	 * {@link DiagramEditorBase}-based, {@code null} otherwise.
	 */
	private EditPartViewer getDiagramEditPartViewer(IHierarchicElement element) {
		IModelEditorBindingService es = IModelEditorBindingService.getInstance();
		es.openInEditor(element);
		while(!es.isOpen(element)) {
			try {
				sleep(100);
			} catch(Exception e) {
				error(ToolingBaseUIActivator.getDefault(), "Error opening editor.", e);
				break;
			}
		}

		IModelEditor<EObject> editor = es.getActiveEditor();
		return editor instanceof DiagramEditorBase ? ((DiagramEditorBase<?>)editor).getViewer()
				: null;

	}

	/**
	 * Creates a KIELER {@link KNode} graph with the same structure as the given
	 * {@link IHierarchicElement}.
	 * 
	 * @param element
	 *            The {@link IHierarchicElement} for which the {@link KNode} graph should be created
	 * 
	 * @return A {@link KNode} graph with the same structure as the graph contained in the
	 *         {@link IHierarchicElement}.
	 */
	private KNode createKIELERGraph(IHierarchicElement element) {

		KNode rootNode = createInitializedNode();

		modelElementsToKNodes = new BasicEMap<IHierarchicElement, KNode>();
		outboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		inboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		connectionsToKEdges = new BasicEMap<IConnection, KEdge>();
		undirectedConnectorsToKPorts = new BasicEMap<IConnector, KPort>();

		EditPartViewer viewer = getDiagramEditPartViewer(element);
		GC gc = viewer != null ? new GC(viewer.getControl()) : null;
		IAutoLayouterTapeMeasure tapeMeasure = new DiagramTapeMeasure(viewer, gc);

		// Create nodes
		for(IHierarchicElement child : element.getContainedElements()) {
			if(!(child instanceof ILayoutedModelElement)) {
				continue;
			}
			// Adjust size of node to number of connectors and label size
			KNode kNode = createInitializedNode();
			KShapeLayout kNodeLayout = kNode.getData(KShapeLayout.class);

			int width = tapeMeasure.getElementWidth(child);
			kNodeLayout.setWidth(truncateSnap2Grid(width, true));
			int height = tapeMeasure.getElementHeight(child);
			kNodeLayout.setHeight(truncateSnap2Grid(height, true));
			kNode.setParent(rootNode);
			modelElementsToKNodes.put(child, kNode);

			// Setup connectors
			EList<IConnector> inputConnectors = new BasicEList<IConnector>();
			EList<IConnector> outputConnectors = new BasicEList<IConnector>();
			EList<IConnector> undirectedConnectors = new BasicEList<IConnector>();
			for(IConnector connector : child.getConnectors()) {
				if(connector instanceof EntryConnectorBase) {
					inputConnectors.add(connector);
				} else if(connector instanceof ExitConnectorBase) {
					outputConnectors.add(connector);
				} else {
					outputConnectors.add(connector);
					undirectedConnectors.add(connector);
				}
			}
			for(IConnector connector : outputConnectors) {
				KPort k = createKPortFromIConnector(connector, kNode, outboundConnectorsToKPorts);
				if(undirectedConnectors.contains(connector)) {
					undirectedConnectorsToKPorts.put(connector, k);
				}
			}
			for(IConnector connector : inputConnectors) {
				createKPortFromIConnector(connector, kNode, inboundConnectorsToKPorts);
			}
		}

		// Create virtual nodes for global ports
		globalInboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		globalOutboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();

		for(IConnector connector : element.getConnectors()) {
			KNode kNodeVirtual = createInitializedNode();
			KShapeLayout kNodeVirtualLayout = kNodeVirtual.getData(KShapeLayout.class);
			kNodeVirtualLayout.setWidth(tapeMeasure.getConnectorWidth(connector));
			kNodeVirtualLayout.setHeight(tapeMeasure.getConnectorHeight(connector));
			kNodeVirtual.setParent(rootNode);

			KPort kPort = createInitializedPort();
			kPort.setNode(kNodeVirtual);
			if(connector instanceof ExitConnectorBase) {
				globalOutboundConnectorsToKPorts.put(connector, kPort);
			} else {
				globalInboundConnectorsToKPorts.put(connector, kPort);
				outboundConnectorsToKPorts.put(connector, kPort);
			}
		}

		// Create edges
		for(Entry<IConnector, KPort> entry : outboundConnectorsToKPorts) {
			for(IConnection connection : entry.getKey().getOutgoing()) {
				KEdge kEdge = createInitializedEdge();
				KPort kSourcePort = entry.getValue();

				kEdge.setSource(kSourcePort.getNode());
				kEdge.setSourcePort(kSourcePort);

				KLabel kLabel = createInitializedLabel(kEdge);
				KShapeLayout kLabelLayout = kLabel.getData(KShapeLayout.class);
				int width = tapeMeasure.getConnectionWidth(connection);
				kLabelLayout.setWidth(truncateSnap2Grid(width, true));

				kSourcePort.getEdges().add(kEdge);
				KPort kTargetPort = inboundConnectorsToKPorts.get(connection.getTarget());
				if(kTargetPort == null) {
					kTargetPort = globalOutboundConnectorsToKPorts.get(connection.getTarget());
				}
				if(kTargetPort == null) {
					kTargetPort = undirectedConnectorsToKPorts.get(connection.getTarget());
				}
				kEdge.setTargetPort(kTargetPort);
				kEdge.setTarget(kTargetPort.getNode());
				kTargetPort.getEdges().add(kEdge);
				connectionsToKEdges.put(connection, kEdge);
			}
		}

		if(gc != null) {
			gc.dispose();
		}
		return rootNode;
	}

	/**
	 * Converts a given {@link IConnector} into a {@link KPort}, and registers it with the given
	 * {@link KNode} (representing the {@link IHierarchicElement} containing the {@link IConnector})
	 * and a {@link IConnector}-to- {@link KPort} map.
	 * 
	 * @param connector
	 *            {@link IConnector} to be converted to {@link KPort}
	 * @param kNode
	 *            {@link KNode} that will contain the newly created {@link KPort}
	 * @param connectorsToKPorts
	 *            Map between {@link IConnector}s and {@link KPort}s.
	 * 
	 * @return {@link KNode} representing the given {@link IConnector}.
	 */
	private KPort createKPortFromIConnector(IConnector connector, KNode kNode,
			EMap<IConnector, KPort> connectorsToKPorts) {
		KPort kPort = createInitializedPort();
		kPort.setNode(kNode);
		KShapeLayout kPortLayout = kPort.getData(KShapeLayout.class);
		kPortLayout.setHeight(DEFAULT_CONNECTOR_SIZE);

		connectorsToKPorts.put(connector, kPort);
		return kPort;
	}
}
